@@ -1,8 +1,8 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "blitzmax-mode" "1.0.0"
+(define-package "blitzmax-mode" "20250221.1408"
   "A major mode for editing BlitzMax source code."
   '((emacs "24.1"))
   :url "https://www.sodaware.net/dev/tools/blitzmax-mode/"
-  :commit "d772deff2464d48d018bbe43b1e4b3745a4ac886"
-  :revdesc "d772deff2464"
+  :commit "adf6444ca4cea047b5532495e9692ed044224ea4"
+  :revdesc "adf6444ca4ce"
   :keywords '("languages" "blitzmax"))
