@@ -2,8 +2,8 @@
 
 ;; Copyright (C) 2012-2020 Phil Newton
 
-;; Package-Version: 1.0.0
-;; Package-Revision: d772deff2464
+;; Package-Version: 20250221.1408
+;; Package-Revision: adf6444ca4ce
 ;; Keywords: languages blitzmax
 ;; Author: Phil Newton
 ;; URL: https://www.sodaware.net/dev/tools/blitzmax-mode/
@@ -35,42 +35,82 @@
 ;; This is a major mode for editing BlitzMax files.  It supports syntax
 ;; highlighting, keyword capitalization, and automatic indentation.
 
-;; If quickrun is installed and enabled, this mode adds BlitzMax for quickrun.
+;; blitzmax-mode provides the following configuration options:
 
-;;; Configuration:
+;; blitzmax-mode-indent
+;;   The number of spaces to indent by.  By default blitzmax-mode indents by 4
+;;   spaces which is converted to a single tab.
 
-(defvar blitzmax-mode-indent 4
-  "Default indentation per nesting level.")
+;; blitzmax-mode-capitalize-keywords-p
+;;   Disable automatic capitalization of keywords by setting this to =nil=.  =t=
+;;   by default.
 
-(defvar blitzmax-mode-smart-indent-p t
-  "Whether to use smart-indentation.")
+;; blitzmax-mode-smart-indent-p
+;;   Disable smart indentation by setting this to =nil=.  =t= by default.
 
-(defvar blitzmax-mode-fontify-p t
-  "Whether to fontify BlitzMax buffers.")
+;; blitzmax-mode-complete-pairs-p
+;;   Enable automatic insertion of matching keyword pairs - such as "End Method"
+;;   - by setting this to =t=.  =nil= by default.
 
-(defvar blitzmax-mode-capitalize-keywords-p t
-  "Whether to automatically capitalize keywords.")
+;; blitzmax-mode-compiler-pathname
+;;   Full pathname to the BlitzMax compiler bmk.  "bmk" by default.
 
-(defvar blitzmax-mode-compiler-pathname "bmk"
-  "The full pathname of the BlitzMax compiler (e.g. /usr/bin/bmk).")
-
-(defvar blitzmax-mode-use-quickrun-p t
-  "Whether to enable quickrun support.")
+;; To enable quickrun integration, add the following code to your Emacs
+;; initialization file.
 
+;;   (with-eval-after-load 'quickrun
+;;     (blitzmax-mode-quickrun-integration))
 
 ;;; Code:
 
+;; --------------------------------------------------
+;; -- Configuration
+
+(defgroup blitzmax nil
+  "Major mode for editing BlitzMax source files."
+  :group 'languages
+  :prefix "blitzmax-")
+
+(defcustom blitzmax-mode-indent 4
+  "Default indentation per nesting level."
+  :type 'integer
+  :group 'blitzmax)
+
+(defcustom blitzmax-mode-smart-indent-p t
+  "Whether to use smart-indentation."
+  :type 'boolean
+  :group 'blitzmax)
+
+(defcustom blitzmax-mode-complete-pairs-p nil
+  "Whether to auto-complete pairs like `Method...End Method'."
+  :type 'boolean
+  :group 'blitzmax)
+
+(defcustom blitzmax-mode-capitalize-keywords-p t
+  "Whether to automatically capitalize keywords."
+  :type 'boolean
+  :group 'blitzmax)
+
+(defcustom blitzmax-mode-compiler-pathname "bmk"
+  "The full pathname of the BlitzMax compiler (e.g. /usr/bin/bmk)."
+  :type 'string
+  :group 'blitzmax)
+
+(defcustom blitzmax-mode-hook nil
+  "Hook run when entering blitzmax mode."
+  :type 'hook
+  :group 'hook)
+
+
 ;; --------------------------------------------------
 ;; -- Local Variables
 
 (defvar blitzmax-mode-abbrev-table nil)
-(defvar blitzmax-mode-hook ())
 (defvar blitzmax-mode-font-lock-keywords)
-(defvar quickrun-file-alist)
 
 
 ;; --------------------------------------------------
-;; -- Regex for highlighting
+;; -- Regex for indentation & completion
 
 (defconst blitzmax-mode-type-start-regexp
   (concat
@@ -88,6 +128,22 @@
 (defconst blitzmax-mode-defun-end-regexp
   "^[ \t]*[Ee]nd[ \t]*\\([Mm]ethod\\|[Ff]unction\\)")
 
+(defconst blitzmax-mode-function-start-regexp
+  (concat
+   "^[\t ]*[Ff]unction"
+   "[ \t ]+_?\\(\\w+\\)[ \t ]*(?"))
+
+(defconst blitzmax-mode-function-end-regexp
+  "^[ \t]*[Ee]nd[ \t]*[Ff]unction")
+
+(defconst blitzmax-mode-method-start-regexp
+  (concat
+   "^[\t ]*[Mm]ethod"
+   "[ \t ]+_?\\(\\w+\\)[ \t ]*(?"))
+
+(defconst blitzmax-mode-method-end-regexp
+  "^[ \t]*[Ee]nd[ \t]*[Mm]ethod")
+
 (defconst blitzmax-mode-abstract-defun-regexp
   (concat
    "^[\t ]*\\([Mm]ethod\\|[Ff]unction\\)"
@@ -234,8 +290,8 @@ a lowercase word, the second is the correctly-capitalized word."
   (let ((all-keywords (append blitzmax-mode-all-keywords
                               blitzmax-mode-type-keywords
                               blitzmax-mode-constant-keywords)))
-    (mapcar #'(lambda (word)
-                (list (downcase word) word))
+    (mapcar (lambda (word)
+              (list (downcase word) word))
             all-keywords)))
 
 (defun blitzmax-mode--create-abbrev-table ()
@@ -255,19 +311,20 @@ Returns `t` if the table was created, `nil` if it already exists."
 Returns `t` if in code, `nil` if in a comment or string."
   (if (fboundp 'buffer-syntactic-context)
       (null (buffer-syntactic-context))
-      (let* ((beg (save-excursion
-                    (beginning-of-line)
-                    (point)))
-             (list
-              (parse-partial-sexp beg (point))))
-        (and (null (nth 3 list))      ;; Is inside string.
-             (null (nth 4 list))))))  ;; Is inside comment.
+    (let* ((beg (save-excursion
+                  (beginning-of-line)
+                  (point)))
+           (list
+            (parse-partial-sexp beg (point))))
+      (and (null (nth 3 list))      ;; Is inside string.
+           (null (nth 4 list))))))  ;; Is inside comment.
 
-(defun blitzmax-mode--capitalize-keywords ()
-  "Automatically capitalize keywords if in a code context."
+(defun blitzmax-mode--capitalize-keywords (expand-fun)
+  "Automatically capitalize keywords if in a code context and call EXPAND-FUN."
   (setq local-abbrev-table
-        (if (blitzmax-mode--in-code-context-p)
-            blitzmax-mode-abbrev-table)))
+        (when (blitzmax-mode--in-code-context-p)
+          blitzmax-mode-abbrev-table))
+  (funcall expand-fun))
 
 
 ;; --------------------------------------------------
@@ -317,13 +374,22 @@ Returns `t` if in code, `nil` if in a comment or string."
 
 (defun blitzmax-mode--previous-line-of-code ()
   "Move to the previous line of code, skipping over any comments or whitespace."
-  (if (not (bobp))
-      (forward-line -1))
+  (unless (bobp)
+    (forward-line -1))
   (while (and (not (bobp))
               (or (looking-at blitzmax-mode-blank-regexp)
                   (looking-at blitzmax-mode-comment-regexp)))
     (forward-line -1)))
 
+(defun blitzmax-mode--next-line-of-code ()
+  "Move to the next line of code, skipping over any comments or whitespace."
+  (unless (eobp)
+    (forward-line 1))
+  (while (and (not (eobp))
+              (or (looking-at blitzmax-mode-blank-regexp)
+                  (looking-at blitzmax-mode-comment-regexp)))
+    (forward-line 1)))
+
 (defun blitzmax-mode--find-original-statement ()
   "Move to original statement if current line is part of a continuation."
   (let ((here (point)))
@@ -523,16 +589,108 @@ Returns `t` if in code, `nil` if in a comment or string."
   (interactive)
   (blitzmax-mode--indent-to-column (blitzmax-mode--calculate-indent)))
 
+(defun blitzmax-mode-newline-and-indent ()
+  "Insert a newline, indent it, and complete any pairs if needed."
+  (interactive)
+  ;; Automatically complete any pairs.
+  (when (and blitzmax-mode-complete-pairs-p
+             (eolp))
+    (blitzmax-mode-insert-matching-pair))
+
+  ;; Insert a newline and indent any code.
+  (newline-and-indent))
+
+
+;; --------------------------------------------------
+;; -- Pair completion
+
+(defun blitzmax-mode--construct-unclosed-p (open-regexp close-regexp)
+  "Search for instances of OPEN-REGEXP before CLOSE-REGEXP.
+
+This is used for constructs that cannot be nested, such as types or methods.  It
+checks if there is an instance of OPEN-REGEXP before CLOSE-REGEXP.  If there
+is, it means the construct is unclosed and can have a pair inserted."
+  (save-excursion
+    ;; TODO: This is pretty darn ugly.
+    (let ((openp   nil)
+          (closedp nil))
+      (while (and (not openp) (not closedp) (not (eobp)))
+        (blitzmax-mode--next-line-of-code)
+        (setq openp   (looking-at open-regexp))
+        (setq closedp (looking-at close-regexp)))
+      (or openp (eobp)))))
+
+(defun blitzmax-mode--can-complete-pair (open-regexp close-regexp)
+  "Check if pair that can be completed.
+
+Checks if point is at the start of the pair (declared via OPEN-REGEXP) and also
+checks if that pair is unclosed (via CLOSE-REGEXP)."
+  (and (looking-at open-regexp)
+       (blitzmax-mode--construct-unclosed-p open-regexp close-regexp)))
+
+(defun blitzmax-mode-insert-matching-pair ()
+  "Complete a pair if the current line is the start."
+  (interactive)
+  ;; TODO: This should only run when at the end of the current line, not the start.
+  (save-excursion
+    (beginning-of-line)
+
+    ;; TODO: This can be cleaned up considerably.
+    (cond ((blitzmax-mode--can-complete-pair blitzmax-mode-extern-start-regexp
+                                             blitzmax-mode-extern-end-regexp)
+           (end-of-line)
+           (insert "\nEnd Extern")
+           (indent-according-to-mode))
+
+          ((blitzmax-mode--can-complete-pair blitzmax-mode-type-start-regexp
+                                             blitzmax-mode-type-end-regexp)
+           (end-of-line)
+           (insert "\nEnd Type"))
+
+          ((and (blitzmax-mode--can-complete-pair blitzmax-mode-function-start-regexp
+                                                  blitzmax-mode-function-end-regexp)
+                (not (blitzmax-mode--abstract-defun-p)))
+           (end-of-line)
+           (insert "\nEnd Function")
+           (indent-according-to-mode))
+
+          ((and (blitzmax-mode--can-complete-pair blitzmax-mode-method-start-regexp
+                                                  blitzmax-mode-method-end-regexp)
+                (not (blitzmax-mode--abstract-defun-p)))
+           (end-of-line)
+           (insert "\nEnd Method")
+           (indent-according-to-mode))
+
+          ;; Nested structures that don't require any close checks.
+          ((looking-at blitzmax-mode-for-regexp)
+           (end-of-line)
+           (insert "\nNext")
+           (indent-according-to-mode))
+
+          ((looking-at blitzmax-mode-while-regexp)
+           (end-of-line)
+           (insert "\nWend")
+           (indent-according-to-mode))
+
+          ((looking-at blitzmax-mode-try-regexp)
+           (end-of-line)
+           (insert "\nEnd Try")
+           (indent-according-to-mode))
+
+          ((looking-at blitzmax-mode-select-regexp)
+           (end-of-line)
+           (insert "\nEnd Select")
+           (indent-according-to-mode)))))
+
 
 ;; --------------------------------------------------
 ;; -- Quickrun Support
 
-(defun blitzmax-mode--setup-quickrun ()
-  "Set up blitzmax-mode with quickrun if not already registered."
-  (when blitzmax-mode-use-quickrun-p
-    (blitzmax-mode--register-quickrun-support)))
+(declare-function quickrun-add-command "quickrun")
+(defvar quickrun-file-alist)
 
-(defun blitzmax-mode--register-quickrun-support ()
+;;;###autoload
+(defun blitzmax-mode-quickrun-integration ()
   "Register BlitzMax with quickrun."
 
   ;; Will compile the current buffer in threaded + debug mode and then run it.
@@ -551,27 +709,19 @@ Returns `t` if in code, `nil` if in a comment or string."
   (add-to-list 'quickrun-file-alist '("\\.bmx$" . "blitzmax")))
 
 
-;; --------------------------------------------------
-;; -- Setup Hooks
-
-;;;###autoload
-(add-to-list 'auto-mode-alist '("\\.bmx\\'" . blitzmax-mode))
-
-(with-eval-after-load 'quickrun
-  (blitzmax-mode--setup-quickrun))
-
-
 ;; --------------------------------------------------
 ;; -- Main Mode
 
 ;;;###autoload
-(define-derived-mode blitzmax-mode fundamental-mode
-  "BlitzMax mode"
+(define-derived-mode blitzmax-mode prog-mode "BlitzMax"
   "Major mode for editing BlitzMax source files."
 
-  ;; Fontify buffer if enabled.
-  (when blitzmax-mode-fontify-p
-    (blitzmax-mode--fontify-buffer))
+  ;; Bind <enter> to custom indentation function.
+  (define-key blitzmax-mode-map (kbd "C-j") #'blitzmax-mode-newline-and-indent)
+  (define-key blitzmax-mode-map (kbd "RET") #'blitzmax-mode-newline-and-indent)
+
+  ;; Fontify buffer.
+  (blitzmax-mode--fontify-buffer)
 
   ;; Enable smart indentation.
   (when blitzmax-mode-smart-indent-p
@@ -580,23 +730,32 @@ Returns `t` if in code, `nil` if in a comment or string."
 
   ;; Enable automatic capitalization.
   (when blitzmax-mode-capitalize-keywords-p
-    (make-local-variable 'pre-abbrev-expand-hook)
-    (add-hook 'pre-abbrev-expand-hook #'blitzmax-mode--capitalize-keywords)
+    (add-hook 'abbrev-expand-functions #'blitzmax-mode--capitalize-keywords nil t)
     (abbrev-mode 1))
 
   ;; Add single line comments to syntax table.
   (modify-syntax-entry ?\' "< b" blitzmax-mode-syntax-table)
   (modify-syntax-entry ?\n "> b" blitzmax-mode-syntax-table)
 
+  ;; Configure comment syntax for comment-line/comment-region.
+  (setq-local comment-start      "' ")
+  (setq-local comment-end        "")
+  (setq-local comment-start-skip "'[ \t]*")
+  (setq-local comment-column     0)
+  (setq-local comment-use-syntax t)
+
   ;; Modify syntax to not treat \ as an escape char. Treat ~ as one instead.
   (modify-syntax-entry ?\\ "." blitzmax-mode-syntax-table)
   (modify-syntax-entry ?~ "\\" blitzmax-mode-syntax-table)
 
   ;; Additional syntax support.
-  (setq syntax-propertize-function blitzmax-mode--syntax-propertize-function)
+  (setq syntax-propertize-function blitzmax-mode--syntax-propertize-function))
 
-  ;; Run hooks.
-  (run-hooks 'blitzmax-mode-hook))
+;;;###autoload
+(add-to-list 'auto-mode-alist '("\\.bmx\\'" . blitzmax-mode))
+
+;;;###autoload
+(add-to-list 'interpreter-mode-alist '("bmx" . blitzmax-mode))
 
 (provide 'blitzmax-mode)
 ;;; blitzmax-mode.el ends here
